@@ -82,12 +82,14 @@ static size_t monitor_jmphash_f (void* obj, size_t size) {
     return ((size_t)mi->id) % size;
 }
 
-static FILE* get_heap_dump_file () {
+static FILE* get_heap_dump_file (char* dumpdir) {
     FILE* fp;
-    snprintf (buf, 256, "jmp_heap_dump_%d", ++heap_dump_id);
+    char buf2[256];
+    snprintf (buf, 256, "%s/jmp_heap_dump_%d", 
+	      (dumpdir == NULL ? "./" : dumpdir), ++heap_dump_id);
     fp = fopen (buf, "w");
-    snprintf (buf, 256, _("createing heap dump file: jmp_heap_dump_%d"), heap_dump_id);
-    set_status (buf);
+    snprintf (buf2, 256, _("createing heap dump file: %s"), buf);
+    set_status (buf2);
     return fp;
 }
 
@@ -371,19 +373,21 @@ void heap_dump_0 (int dump_level,
     check_objects (c);
 }
 
-void heap_dump_1 (int dump_level,
-		  char *begin,
-		  char *end,
+void heap_dump_1 (char* dumpdir,
+		  int dump_level,
+		  char* begin,
+		  char* end,
 		  jint num_traces,
 		  JVMPI_CallTrace *traces) {
-    heap_dump_2 (dump_level, begin, end, num_traces, traces);
+    heap_dump_2 (dumpdir, dump_level, begin, end, num_traces, traces);
 }
 
-void heap_dump_2 (int dump_level,
-		  char *begin,
-		  char *end,
+void heap_dump_2 (char* dumpdir,
+		  int dump_level,
+		  char* begin,
+		  char* end,
 		  jint num_traces,
-		  JVMPI_CallTrace *traces) {
+		  JVMPI_CallTrace* traces) {
 
     unsigned char type, ty;
     
@@ -396,7 +400,7 @@ void heap_dump_2 (int dump_level,
     size_t counter = 0;
     char* p = begin;    
     size_t instances = 0;
-    FILE* fp = get_heap_dump_file ();
+    FILE* fp = get_heap_dump_file (dumpdir);
     
     fprintf (fp, "heap_dump: %d, %p, %p, %d, %p\n", 
 	     dump_level, begin, end, num_traces, traces);
@@ -541,8 +545,10 @@ void heap_dump_2 (int dump_level,
 		     oid, clz, (c ? c->name : "?"), u4);
 	    while (c != NULL) {
 		jobjectID sobj;
-		fprintf (fp, "class '%s' has %d variables...\n", 
-			 c->name, c->n_instances);
+		if (c->n_instances > 0) {
+		    fprintf (fp, "class '%s' has %d variables...\n", 
+			     c->name, c->n_instances);
+		}
 		for (i = 0; i < c->n_instances; i++) {
 		    sobj = get_variable (fp, &p, &c->instances[i]);
 		    if (sobj)
@@ -557,7 +563,6 @@ void heap_dump_2 (int dump_level,
 		fprintf (stderr, "doh, failed to read instance variables size: %d.....\n", u4);
 		p = pe;
 	    }
-	    fprintf (fp, "instance done...\n");
 	    break;
 	}
 	case JVMPI_GC_OBJ_ARRAY_DUMP:
