@@ -24,6 +24,7 @@
 #include <filter.h>
 #include <jmptime.h>
 #include <deadlock_detector.h>
+#include <string_dumper.h>
 
 #ifdef linux
 /* back trace handling only works in linux for what I know. */
@@ -48,10 +49,14 @@ static JVMPI_Interface* jvmpi;
 
 /** are we showing any graphical user interface? */
 static int doUI = 1;
+/** are we profiling classes. */
+static int class_profiling = 0;
 /** are we profiling method calls? */
 static int method_profiling = 1;
 /** are we tracing objects? */
 static int object_profiling = 1;
+/** Do object allocations follow the filter? */
+static int alloc_follow_filter = 0;
 /** are we tracing monitors. */
 static int monitor_profiling = 1;
 /** do we allow dump on signal? */
@@ -75,9 +80,11 @@ static hashtab* methods;
 static hashtab* threads;
 
 /** The current reset level. */
-static int current_level = 0;
+static int current_reset_level = 0;
 /** The minimum level needed when getting objects. */
-static int minimum_level = 0;
+static int minimum_reset_level = 0;
+/** The current gc level. */
+static int current_gc_level = 0; 
 
 /** This is our objectstore. */
 static objectstore* osp;
@@ -120,13 +127,16 @@ static jlong this_gc_time = 0;
 static jlong heap_size = 0;
 
 /** Where should we dump data files? */
-static char* dumpfile = NULL;
+static char* dumpdir = NULL;
 
 static int down = 0;   /** since we get multiple shutdowns... */
 JVMPI_RawMonitor ui_shutdown_lock;
 
 static hashtab* last_monitor_dump;
 
+/** we need to make sure that we call the right set_status during gc.*/
+static int gc_requested = 0;
+
 #ifdef linux
 void get_backtrace () {
     int i;
@@ -188,6 +198,11 @@ inline int tracing_objects () {
     return object_profiling;
 }
 
+/** Test if object allocations follows filter. */
+inline int allocs_follow_filter () {
+    return alloc_follow_filter;
+}
+
 /** Test if jmp is currently tracing method calls. */
 inline int tracing_methods () {
     return method_profiling;
@@ -207,8 +222,8 @@ inline int usingUI () {
 void reset_counters () {
     jmphash_lock (classes);
     jmphash_for_each ((jmphash_iter_f)cls_reset_count, classes);
-    current_level++;
-    minimum_level = current_level;
+    current_reset_level++;
+    minimum_reset_level = current_reset_level;
     jmphash_unlock (classes);
     jmphash_lock (methods);
     jmphash_for_each ((jmphash_iter_f)method_reset_count, methods);
@@ -219,7 +234,7 @@ void reset_counters () {
 /** Restore the counter to theire full values. */
 void restore_counters () {
     jmphash_lock (classes);    
-    minimum_level = 0;
+    minimum_reset_level = 0;
     jmphash_for_each ((jmphash_iter_f)cls_restore_count, classes);
     jmphash_unlock (classes);
     jmphash_lock (methods);
@@ -231,7 +246,9 @@ static void object_free (jobjectID);
 
 /** Run the garbage collector. */
 void run_GC () {
+    gc_requested = 1;
     jvmpi->RunGC ();    
+    gc_requested = 0;
 }
 
 /** Get the size of the heap. */
@@ -252,6 +269,36 @@ static void free_the_object (void* v) {
     objectstore_obj_free (osp, op);
 }
 
+static void release_object (obj* o) {
+    objectstore_obj_free (osp, o);
+}
+
+/** Load a primitive array class. */
+static cls* load_array_class (char* cls_name) {
+    cls* c;
+    c = cls_new (cls_name, cls_name, (jobjectID)cls_name, 0, 0, NULL, 0, NULL);
+    if (classes) {
+	if (c)
+	    jmphash_insert (c, classes);
+	else 
+	    fprintf (stderr, "failed to allocate array class: '%s'\n", cls_name);
+    }
+    return c;
+}
+
+/** Load all of the primitive array classes. */
+static void load_array_classes () {
+    LP = load_array_class (L);
+    ZP = load_array_class (Z);
+    BP = load_array_class (B);
+    CP = load_array_class (C);
+    SP = load_array_class (S);
+    IP = load_array_class (I);
+    JP = load_array_class (J);
+    FP = load_array_class (F);
+    DP = load_array_class (D);
+}
+
 /** Perform a heao dump with the specified level.
  */
 static void run_heap_dump_arg (int level) {
@@ -266,6 +313,17 @@ static void run_heap_dump_arg (int level) {
     fprintf (stderr, "requesting heap dump: %d\n", level);
     jvmpi->RequestEvent (JVMPI_EVENT_HEAP_DUMP, &heap_arg);
     fprintf (stderr, "heap dumping done: %d\n", level);
+    if (class_profiling == 0) {
+	/* clear classes, objects and methods... */
+	jmphash_for_each ((jmphash_iter_f)release_object, objects);
+	jmphash_clear (objects);
+	jmphash_for_each ((jmphash_iter_f)method_free, methods); 
+	jmphash_clear (methods);
+	jmphash_for_each ((jmphash_iter_f)cls_free, classes); 
+	jmphash_clear (classes);
+	/* and fill in with standard information... */
+	load_array_classes ();    
+    }
     unlock_all ();
 }
 
@@ -277,6 +335,10 @@ void run_object_dump () {
     run_heap_dump_arg (JVMPI_DUMP_LEVEL_0);
 }
 
+void run_string_dump () {
+    dump_strings ();
+}
+
 hashtab* run_monitor_dump () {
     jvmpi->RequestEvent (JVMPI_EVENT_MONITOR_DUMP, NULL);
     detect_deadlock (last_monitor_dump);
@@ -330,32 +392,6 @@ void get_arena_load (jint arena_id) {
 		 arena_id, res);
 }
 
-/** Load a primitive array class. */
-static cls* load_array_class (char* cls_name) {
-    cls* c;
-    c = cls_new (cls_name, cls_name, (jobjectID)cls_name, 0, 0, NULL, 0, NULL);
-    if (classes) {
-	if (c)
-	    jmphash_insert (c, classes);
-	else 
-	    fprintf (stderr, "failed to allocate array class: '%s'\n", cls_name);
-    }
-    return c;
-}
-
-/** Load all of the primitive array classes. */
-static void load_array_classes () {
-    LP = load_array_class (L);
-    ZP = load_array_class (Z);
-    BP = load_array_class (B);
-    CP = load_array_class (C);
-    SP = load_array_class (S);
-    IP = load_array_class (I);
-    JP = load_array_class (J);
-    FP = load_array_class (F);
-    DP = load_array_class (D);
-}
-
 /** Do initial setup. */
 static int setup () {
     trace ("JMP setup ()\n");
@@ -436,8 +472,8 @@ static int teardown () {
     free_and_cleanup_hash (&arenas, (jmphash_iter_f)arena_free);
     free_and_cleanup_hash (&threads, (jmphash_iter_f)jthread_free);
     fprintf (stdout, _("done freeing jmp-data..\n"));    
-    free (dumpfile); /* NULL or malloc'ed */
-    dumpfile = NULL;
+    free (dumpdir); /* NULL or malloc'ed */
+    dumpdir = NULL;
     fprintf (stdout, _("teardown complete.\n"));
     return 0;
 }
@@ -529,6 +565,10 @@ static void jvm_init_done (JNIEnv* env) {
     trace ("jvm_init_done.\nContinuing with JMP setup...\n");
     
     trace ("Going to evaluate vm version\n");
+    if (env == NULL) {
+	fprintf (stderr, "Warning: jvm_init_done with JNIEnv == NULL, assuming simulator\n");
+	simulator = 1;
+    }
     if (!simulator) {
 	c_system = (*env)->FindClass (env, "java/lang/System");
 	m_get_property =
@@ -556,16 +596,16 @@ static void jvm_init_done (JNIEnv* env) {
 		     tstatus, 
 		     JNI_OK, 
 		     JNI_ERR);
-    }
+    } 
     
     trace ("JMP setup complete...\n");    
     
-}
+} 
 
 static int run_data_dump_unlocked () {
     int ret = -1;
     if (objects && classes && methods) 
-	ret = dump_data (dumpfile, classes, methods, threads);
+	ret = dump_data (dumpdir, classes, methods, threads);
     return ret;
 }
 
@@ -584,7 +624,7 @@ void
 run_show_objects_alloced_by_method (method* m) {
     trace2 ("run_show_objects_alloced_by_method (%s)\n", m->jmpname);
     if (!down && objects)
-	show_objects_alloced_by_method (objects, m, minimum_level);
+	show_objects_alloced_by_method (objects, m, minimum_reset_level);
 }
 
 
@@ -592,7 +632,7 @@ void
 run_show_objects_alloced_by_class (cls* c) {
     trace2 ("run_show_objects_alloced_by_class (%s)\n", c->name);
     if (!down && objects)
-	show_objects_alloced_by_class (objects, c, minimum_level);
+	show_objects_alloced_by_class (objects, c, minimum_reset_level);
 }
 
 
@@ -622,7 +662,7 @@ typedef struct string_callback_data {
 } string_callback_data;
 
 static void filter_strings (obj* o, string_callback_data* scd) {
-    if (obj_get_class (o) == CP && minimum_level <= obj_get_level (o))
+    if (obj_get_class (o) == CP && minimum_reset_level <= obj_get_reset_level (o))
 	scd->sc (o, scd->data);
 }
 
@@ -738,6 +778,10 @@ jint get_thread_state (const jthread* jt) {
     return jvmpi->GetThreadStatus (jt->env_id);
 }
 
+int get_current_gc_level () {
+    return current_gc_level;
+}
+
 static void disable_events (int keep_threads) {
     /* Disable standard events */
     jvmpi->DisableEvent (JVMPI_EVENT_JVM_INIT_DONE, NULL);
@@ -794,15 +838,17 @@ static void jvm_shut_down () {
     */
 }
 
-
 /** grab the locks needed during gc.
  */
-static void gc_start() {
-    trace ("gc_start ()\n");
+static void gc_start () {
+    trace2 ("gc_start (), %d\n", gc_requested);
     if (down)
 	return;
-    set_status (_("Running garbage collection..."));
     lock_all ();
+    if (gc_requested)
+	set_status (_("Running garbage collection..."));
+    else 
+	set_status_lock (_("Running garbage collection..."));
     this_gc_object_move = 0;
     this_gc_object_free = 0;
     this_gc_time = get_absolute_time (jvmpi);
@@ -821,14 +867,18 @@ static void gc_finish (jlong used_objects,
     if (down)
 	return;
     remove_owners_information ();
-    trace ("gc_finish ()\n");
-    unlock_all ();
+    trace2 ("gc_finish (), %d\n", gc_requested);
     snprintf (buf, 200, 
 	      _("Garbage collection completed: %ld objects moved, "
 		"%ld objects freed in %lld.%06lld seconds"), 
 	      this_gc_object_move, this_gc_object_free, jsec, jnsec);
-    set_status (buf);
+    if (gc_requested)
+	set_status (buf); 
+    else 
+	set_status_lock (buf); 
     heap_size = total_object_space;
+    current_gc_level++;
+    unlock_all ();
 }
 
 
@@ -841,7 +891,7 @@ static void jvmpi_arena_new (jint arena_id, const char* arena_name) {
     snprintf (buffer, 80, _("Arena %d, %s created."), 
 	      arena_id, arena_name);
     fprintf (stderr, "arena_new: %d, %s\n", arena_id, arena_name);
-    set_status (buffer);
+    set_status_lock (buffer);
     trace3 ("arena_new: (%d, %s)\n", arena_id, arena_name);
     if (!arenas)
 	return;
@@ -1088,7 +1138,7 @@ static void object_alloc (jint arena_id,
     jmphash_lock (classes);
     cp = get_class_pointer (is_array, class_id);
     if (cp) 
-	cls_object_alloc (cp, size);
+	cls_object_alloc (cp, size, current_gc_level);
     jmphash_unlock (classes);
 
     /* We dont change cp from now on,
@@ -1115,7 +1165,10 @@ static void object_alloc (jint arena_id,
 		timerstack_lock (t);
 		if (t->top > 0) {
 		    mt = t->times + (t->top - 1);
-		    mp = mt->method;
+		    if (allocs_follow_filter ())
+			mp = mt->filtered_method;
+		    else 
+			mp = mt->method;
 		    mp->allocated_objects++;
 		    mp->allocated_memory += size;
 		    mp->modified = 1;
@@ -1129,7 +1182,8 @@ static void object_alloc (jint arena_id,
 	 * objectstore_obj_new is thread safe 
 	 */
 	op = objectstore_obj_new (osp, arena_id, cp, is_array, size, 
-				  obj_id, mp, current_level);
+				  obj_id, mp, current_reset_level, 
+				  current_gc_level);
 	jmphash_insert (op, objects);
 	jmphash_unlock (objects);
     } else {
@@ -1191,7 +1245,6 @@ static void object_free (jobjectID obj_id) {
     obj* op = NULL;
     obj* hop = NULL;
     cls* cp = NULL;
-    jint os;
     
     trace2 ("object_free (%p)\n", obj_id);
     this_gc_object_free++;
@@ -1204,8 +1257,9 @@ static void object_free (jobjectID obj_id) {
 	hop = jmphash_remove (op, objects);
 	cp = get_class (obj_get_class_id (hop));
 	if (cp != NULL) {
-	    os = obj_get_size (hop);
-	    cls_object_free (cp, os);
+	    jint os = obj_get_size (hop);
+	    jint ol = obj_get_gc_level (hop);
+	    cls_object_free (cp, os, ol);
 	}
 	objectstore_obj_free (osp, hop);
     }
@@ -1250,7 +1304,7 @@ static void thread_start (char* thread_name,
     char buffer[80];
     
     snprintf (buffer,80, _("Thread %s started"), thread_name);
-    set_status (buffer);
+    set_status_lock (buffer);
     trace4 ("thread_start (%s,%s,%s,", thread_name, group_name, parent_name);
     trace3 ("%p,%p)\n", thread_id, thread_env_id);
     jmphash_lock (threads);
@@ -1315,8 +1369,11 @@ static method* get_unknown_method (jmethodID method_id) {
     tried to get unknown method: 0xca4b9a0 => class: 0x90099e8 
         (sun.reflect.GeneratedSerializationConstructorAccessor122) => (nil)
     */
-    fprintf (stderr, "tried to get unknown method: %p => class: %p (%s) => %p\n", 
-	     method_id, c, (c ? cls_get_name (c) : "?"), m);
+    if (m == NULL) {
+	fprintf (stderr, 
+		 "tried to get unknown method: %p => class: %p (%s) => %p\n",
+		 method_id, c, (c ? cls_get_name (c) : "?"), m);
+    }
     return m;
 }
 
@@ -1366,10 +1423,10 @@ static void method_exit (jmethodID method_id, JNIEnv* env) {
 
 
 static void heap_dump (int dump_level,
-		       char *begin,
-		       char *end,
+		       char* begin,
+		       char* end,
 		       jint num_traces,
-		       JVMPI_CallTrace *traces) {
+		       JVMPI_CallTrace* traces) {
     trace4 ("heap dump dump_level: %d, begin: %p, end: %p, ", dump_level, begin, end);
     trace3 ("num_traces: %d, traces: %p\n", num_traces, traces);    
     switch (dump_level) {
@@ -1377,10 +1434,10 @@ static void heap_dump (int dump_level,
 	heap_dump_0 (dump_level, begin, end, num_traces, traces);
 	break;
     case JVMPI_DUMP_LEVEL_1:
-	heap_dump_1 (dump_level, begin, end, num_traces, traces);
+	heap_dump_1 (dumpdir, dump_level, begin, end, num_traces, traces);
 	break;
     case JVMPI_DUMP_LEVEL_2:
-	heap_dump_2 (dump_level, begin, end, num_traces, traces);
+	heap_dump_2 (dumpdir, dump_level, begin, end, num_traces, traces);
 	break;
     }
 }
@@ -1628,9 +1685,10 @@ static void display_help () {
     fprintf (stdout, "nomethods - to disable method profiling.\n");
     fprintf (stdout, "noobjects - to disable object profiling.\n");
     fprintf (stdout, "nomonitors - to disable monitor profiling.\n");
+    fprintf (stdout, "allocfollowsfilter - to group object allocations into filtered methods.\n");    
     fprintf (stdout, "nogui     - to run jmp without the user interface.\n");
     fprintf (stdout, "dodump    - to allow to be called with signals.\n");
-    fprintf (stdout, "dumpfile=<filename> - to specify the dumpfiles.\n");
+    fprintf (stdout, "dumpdir=<directoryr> - to specify where the dump-/heapdumpfiles go.\n");
     fprintf (stdout, "dumptimer=<n> - to specify automatic dump every n:th second.\n");
     fprintf (stdout, "filter=<somefilter> - to specify an initial recursive filter.\n");
     fprintf (stdout, "threadtime - to specify that timing of methods and monitors \n"
@@ -1638,7 +1696,7 @@ static void display_help () {
     fprintf (stdout, "simulator - to specify that jmp should not perform any jni tricks.\n"
 	     "           probably only useful if you debug jmp.\n");
     fprintf (stdout, "\nAn example may look like this:\n"); 
-    fprintf (stdout, "java -Xrunjmp:nomethods,dumpfile=/tmp/jmpdump rabbit.proxy.Proxy\n"); 
+    fprintf (stdout, "java -Xrunjmp:nomethods,dumpdir=/tmp/jmpdump/ rabbit.proxy.Proxy\n"); 
 }
 
 static size_t get_size_of_option (char* option) {
@@ -1738,12 +1796,12 @@ static void parse_options (char* options) {
 	m = strstr (options, "simulator");
 	simulator = (m != NULL);
 	
-	m = strstr (options, "dumpfile=");
+	m = strstr (options, "dumpdir=");
 	if (m != NULL) {
-	    size_t t = get_size_of_option (m + 9);	    
-	    dumpfile = malloc (t + 1);
-	    strncpy (dumpfile, m + 9, t);
-	    dumpfile[t] = '\0';
+	    size_t t = get_size_of_option (m + 8);	    
+	    dumpdir = malloc (t + 1);
+	    strncpy (dumpdir, m + 8, t);
+	    dumpdir[t] = '\0';
 	}
 	
 	m = strstr (options, "dumptimer=");
@@ -1758,6 +1816,11 @@ static void parse_options (char* options) {
 		 set_dump_timer (l);
 	    }		
 	}
+
+	m = strstr (options, "allocfollowsfilter");
+	if (m != NULL) {
+	    alloc_follow_filter = 1;
+	}
 	
 	co = options;
 	while (co) {
@@ -1771,7 +1834,7 @@ static void parse_options (char* options) {
 	
 	absolute_times = strstr (options, "threadtime") == NULL;
     } else {
-	// to be consistent with above...
+	/* to be consistent with above... */
 	absolute_times = 1;
     }
 }
@@ -1787,13 +1850,20 @@ static void enable_arena_events () {
 }
 
 static void enable_class_events () {
+    class_profiling = 1;
     jvmpi->EnableEvent (JVMPI_EVENT_CLASS_LOAD, NULL); 
     jvmpi->EnableEvent (JVMPI_EVENT_CLASS_UNLOAD, NULL);
     /* classes are moved so we need this here.. */
     jvmpi->EnableEvent (JVMPI_EVENT_OBJECT_MOVE, NULL);
 }
 
-void enable_object_events () {   
+static void ensure_class_profiling () {
+    if (class_profiling == 0)
+	enable_class_events ();
+}
+
+void enable_object_events () {
+    ensure_class_profiling ();
     object_profiling = 1;
     /* Enable object tracing events. */
     jvmpi->EnableEvent (JVMPI_EVENT_OBJECT_ALLOC, NULL);
@@ -1839,7 +1909,7 @@ void get_call_trace_env (JNIEnv* tid) {
 	ts->top = 0;
 	ts->last_contentation = -1;
 	jvmpi->GetCallTrace (&ct, depth);
-	fprintf (stderr, "got: %d methods for %p...\n", ct.num_frames, tid); 
+	trace3 ("got: %d methods for %p...\n", ct.num_frames, tid); 
 	for (i = ct.num_frames - 1; i >= 0; i--) {
 	    method* m;
 	    JVMPI_CallFrame* f = ct.frames + i;
@@ -1860,30 +1930,43 @@ void get_call_trace_env (JNIEnv* tid) {
 }
 
 static void enable_method_events_and_stacks (int get_stacks) {
+    ensure_class_profiling ();
+
     /* Ok, the thread stacks are bogous, clear them all and try to get 
      * real stacks, timing will be a bit bogous, but not to much...
      */
     
     if (get_stacks) {
+	int locks;
 	jmphash_lock (threads);
 	jmphash_lock (methods);
+	locks = timerstacks_get_need_locks ();
+	timerstacks_set_need_locks (1);
+
+	/*
+	  WARNING! due to a bug inside SUN's jvm
+	  method profiling may cause the jvm to crash see:
+	  http://developer.java.sun.com/developer/bugParade/bugs/4652208.html
+	  for more info
+	*/
+
+	/* Enable method profiling events under lock, so we 
+	 * don't get odd method entries. 
+	 */
+	jvmpi->EnableEvent (JVMPI_EVENT_METHOD_ENTRY, NULL);
+	jvmpi->EnableEvent (JVMPI_EVENT_METHOD_EXIT, NULL);
+	
 	jvmpi->DisableGC ();
 	jmphash_for_each ((jmphash_iter_f)get_call_trace, threads);
 	jvmpi->EnableGC ();
+	timerstacks_set_need_locks (locks);
 	jmphash_unlock (methods);
 	jmphash_unlock (threads);
+    } else {
+	jvmpi->EnableEvent (JVMPI_EVENT_METHOD_ENTRY, NULL);
+	jvmpi->EnableEvent (JVMPI_EVENT_METHOD_EXIT, NULL);
     }
-
     method_profiling = 1;
-    /*
-      WARNING! due to a bug inside SUN's jvm
-      method profiling may cause the jvm to crash see:
-      http://developer.java.sun.com/developer/bugParade/bugs/4652208.html
-      for more info
-    */
-    /* Enable method profiling events. */
-    jvmpi->EnableEvent (JVMPI_EVENT_METHOD_ENTRY, NULL);
-    jvmpi->EnableEvent (JVMPI_EVENT_METHOD_EXIT, NULL);
 }
 
 void enable_method_events () {
@@ -1914,7 +1997,10 @@ static void enable_events () {
     jvmpi->EnableEvent (JVMPI_EVENT_JVM_SHUT_DOWN, NULL);    
     enable_gc_events ();
     enable_arena_events ();
-    enable_class_events ();
+    /* we dont have to enable class events yet, object 
+     * and/or method profiling will do that for us.
+     * enable_class_events ();
+     */
     if (object_profiling)
 	enable_object_events ();
     enable_thread_events ();
@@ -1964,6 +2050,9 @@ JNIEXPORT jint JNICALL JVM_OnLoad (JavaVM *jvm, char *options, void *reserved) {
     /* initialize jvmpi interface */
     enable_events ();
 
+    /* initialize the ui */
+    init_ui ();
+
     return JNI_OK;    
 }
 
