@@ -32,6 +32,7 @@ enum {
     OD_ARENA,
     OD_IS_ARRAY,
     OD_SIZE,
+    OD_TENURE,
     OD_ALLOC_CLASS,
     OD_ALLOC_METHOD,
     OD_VARIABLE,
@@ -113,6 +114,17 @@ static gint sort_by_size (GtkTreeModel *model,
     return o1->size - o2->size;
 }
 
+static gint sort_by_tenure (GtkTreeModel *model,
+			    GtkTreeIter *a,
+			    GtkTreeIter *b,
+			    gpointer user_data) {
+    obj* o1;
+    obj* o2;
+    gtk_tree_model_get (model, a, OD_OBJECT, &o1, -1);
+    gtk_tree_model_get (model, b, OD_OBJECT, &o2, -1); 
+    return obj_get_gc_level (o2) - obj_get_gc_level (o1);
+}
+
 static gint sort_by_alloc_class_and_method (GtkTreeModel *model,
 					    GtkTreeIter *a,
 					    GtkTreeIter *b,
@@ -161,6 +173,7 @@ static compr_func obj_comprs[] = {
     sort_by_arena, 
     sort_by_is_array,          /* is_array */
     sort_by_size,              /* size */
+    sort_by_tenure, 
     sort_by_alloc_class_and_method,  /* alloc_class */
     sort_by_alloc_class_and_method,   /* alloc_method */
     sort_by_variable                  /* variable */
@@ -199,6 +212,7 @@ static void append_object (GtkTreeStore* list, GtkTreeIter* iter, GtkTreeIter* p
 			OD_ARENA, text[1],
 			OD_IS_ARRAY, o->is_array,
 			OD_SIZE, obj_get_size (o),
+			OD_TENURE, get_current_gc_level () - obj_get_gc_level (o), 
 			OD_ALLOC_CLASS, text[4],
 			OD_ALLOC_METHOD, text[5],
 			OD_VARIABLE, text[OD_VARIABLE],
@@ -227,6 +241,7 @@ static void append_string (GtkTreeStore* list, GtkTreeIter* iter, GtkTreeIter* p
     text[OD_ARENA] = "";
     text[OD_IS_ARRAY] = "";
     text[OD_SIZE] = "";
+    text[OD_TENURE] = "";
     text[OD_ALLOC_CLASS] = "";
     text[OD_ALLOC_METHOD] = "";
     text[OD_VARIABLE] = "";
@@ -234,11 +249,12 @@ static void append_string (GtkTreeStore* list, GtkTreeIter* iter, GtkTreeIter* p
     gtk_tree_store_append (list, iter, parent);
     gtk_tree_store_set (list, iter,
 			OD_CLASS, text[OD_CLASS],
-			OD_ARENA, text[1],
+			OD_ARENA, text[OD_ARENA],
 			OD_IS_ARRAY, 1,
 			OD_SIZE, len,
-			OD_ALLOC_CLASS, text[4],
-			OD_ALLOC_METHOD, text[5],
+			OD_TENURE, 0,
+			OD_ALLOC_CLASS, text[OD_ALLOC_CLASS],
+			OD_ALLOC_METHOD, text[OD_ALLOC_METHOD],
 			OD_VARIABLE, text[OD_VARIABLE],
 			OD_OBJECT, NULL,
 			-1);
@@ -252,13 +268,13 @@ static void append_object_dump (obj* o, wret* w) {
 
 static void print_if_method (obj* o, wret* w) {
     if (o->method == (method*)w->data && 
-	w->level <= obj_get_level (o))
+	w->level <= obj_get_reset_level (o))
 	append_object_dump (o, w);
 }
 
 static void print_if_class (obj* o, wret* w) {
     if (obj_get_class_id (o) == ((cls*)w->data)->class_id && 
-	w->level <= obj_get_level (o))
+	w->level <= obj_get_reset_level (o))
 	append_object_dump (o, w);
 }
 
@@ -298,36 +314,20 @@ static char* get_field (down_link* dl) {
     return buf;
 }
 
-static void expand_node (GtkTreeStore* treestore, GtkObject* list, GtkTreeIter* node) {
-    GtkTreePath* path;
-    path = gtk_tree_model_get_path (GTK_TREE_MODEL (treestore), node);
-    gtk_tree_view_expand_row (GTK_TREE_VIEW (list), path, FALSE); 
-    gtk_tree_path_free (path);			
-}
-
 /** request to show all live objects by a certain class. */
 static void inspect_instance (GtkObject* list) {
     if (last_selected_object != NULL) {
-	GtkTreeIter iter, tmp;
+	GtkTreeIter iter;
 	GtkTreeIter parent;
 	GtkTreeModel *model;
 	GtkTreeSelection *selection;
-	gboolean valid;
 	down_link* dl;
 	get_instance_info (last_selected_object);
 	dl = get_last_down_link ();
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (list));
 	if (gtk_tree_selection_get_selected (selection, &model, &parent)) {
 	    GtkTreeStore *treestore = (GtkTreeStore*)gtk_tree_view_get_model (GTK_TREE_VIEW (list));
-	    if (gtk_tree_model_iter_has_child (GTK_TREE_MODEL (treestore), &parent)) {
-		gtk_tree_model_iter_children (GTK_TREE_MODEL (treestore), &iter, &parent);
-		valid = TRUE;
-		while (valid) {
-		    memcpy (&tmp, &iter, sizeof (GtkTreeIter));
-		    valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (treestore), &iter);
-		    gtk_tree_store_remove (treestore, &tmp);
-		}
-	    }
+	    remove_children (treestore, &parent);
 	    while (dl != NULL) {
 		switch (dl->type) {
 		case JVMPI_NORMAL_OBJECT: {
@@ -408,7 +408,8 @@ static int still_untraced (obj* o, hashtab* traced) {
     return (v == NULL);
 }
 
-static void update_statistics (hashtab* statistics, obj* o) {
+static void update_statistics (hashtab* statistics, obj* o, 
+			       struct class_statistic* total) {
     struct class_statistic* csr = NULL;
     struct class_statistic* cs = calloc (1, sizeof (*cs));
     cs->clz = obj_get_class (o);
@@ -416,16 +417,21 @@ static void update_statistics (hashtab* statistics, obj* o) {
     if (csr != NULL) {
 	csr->count++;
 	csr->byte_usage += obj_get_size (o);
+	total->count++;
+	total->byte_usage += obj_get_size (o);
 	free (cs);
     } else {
 	cs->count = 1; 
 	cs->byte_usage = obj_get_size (o);
+	total->count++;
+	total->byte_usage += obj_get_size (o);
 	jmphash_insert (cs, statistics);
     }   
 }
 
 static void handle_object (obj* o, struct obj_list* current,
-			   hashtab* traced, hashtab* statistics) {
+			   hashtab* traced, hashtab* statistics, 
+			   struct class_statistic* total) {
     if (o != NULL) {
 	if (still_untraced (o, traced)) {
 	    struct obj_list* ut;
@@ -439,7 +445,7 @@ static void handle_object (obj* o, struct obj_list* current,
 	    ut->next = current->next;
 	    current->next = ut;
 			    
-	    update_statistics (statistics, o);
+	    update_statistics (statistics, o, total);
 	}
     }
 
@@ -535,6 +541,7 @@ static void list_all_owned_objects (GtkObject* list) {
     if (last_selected_object != NULL) {
 	down_link* dl;
 	struct obj_list* untraced;  /* list of objects to trace */       
+	struct class_statistic* total; /* a sum of all data */
 	hashtab* statistics = 
 	    jmphash_new (42, cls_statistic_jmphash_func,
 			 cls_statistic_cmp_func, _("statistics"));
@@ -543,8 +550,12 @@ static void list_all_owned_objects (GtkObject* list) {
 	   marked for tracing. */
 	hashtab* traced = jmphash_new (42, obj_jmphash_func, 
 				       obj_cmp_func, _("traced"));
+
+	total = calloc (1, sizeof (*total));
+	total->clz = cls_new (_("Total"), _("Total"), NULL, 0, 0, NULL, 0, NULL);
+	jmphash_insert (total, statistics);
 	
-	update_statistics (statistics, last_selected_object);
+	update_statistics (statistics, last_selected_object, total);
 
 	/* avoid circular objects... */
 	jmphash_insert (last_selected_object, traced);	
@@ -568,7 +579,7 @@ static void list_all_owned_objects (GtkObject* list) {
 		    /* fallthrough */
 		case JVMPI_CLASS: {
 		    obj* o = get_object (dl->value.o);
-		    handle_object (o, current, traced, statistics);
+		    handle_object (o, current, traced, statistics, total);
 		    break;
 		}
 		}
@@ -580,6 +591,7 @@ static void list_all_owned_objects (GtkObject* list) {
 	}
 
 	show_statistics_window (statistics);
+	cls_free (total->clz);
 	jmphash_for_each ((jmphash_iter_f)free_statistics_object, statistics);
 	jmphash_free (statistics);
 	jmphash_free (traced);
@@ -635,13 +647,15 @@ wret create_object_dump_window () {
     scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
     gtk_container_add (GTK_CONTAINER (obj_dump), scrolledwindow);
     clist1 = gtk_tree_store_new (ODN_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, 
-				 G_TYPE_BOOLEAN, G_TYPE_LONG, G_TYPE_STRING, 
-				 G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);
+				 G_TYPE_BOOLEAN, G_TYPE_LONG, G_TYPE_LONG, 
+				 G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, 
+				 G_TYPE_POINTER);
     tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (clist1));
     add_column (tree, _("Class"), OD_CLASS, (gpointer)tree, od_column_clicked, 200, 0);
     add_column (tree, _("Arena"), OD_ARENA, (gpointer)tree, od_column_clicked, 80, 0);
     add_column (tree, _("Is array"), OD_IS_ARRAY, (gpointer)tree, od_column_clicked, 80, 0);
     add_column (tree, _("Size"), OD_SIZE, (gpointer)tree, od_column_clicked, 80, 0);
+    add_column (tree, _("Tenure"), OD_TENURE, (gpointer)tree, od_column_clicked, 80, 0);
     add_column (tree, _("Allocating class"), OD_ALLOC_CLASS, (gpointer)tree, od_column_clicked, 200, 0);
     add_column (tree, _("Allocating method"), OD_ALLOC_METHOD, (gpointer)tree, od_column_clicked, 200, 0);
     add_column (tree, _("Variable"), OD_VARIABLE, (gpointer)tree, od_column_clicked, 200, 0);
