@@ -10,10 +10,14 @@ enum {
     OINSTANCE_MAX_COLUMN,
     OSIZE_COLUMN,
     OGC_COLUMN,
+    OTENURE_COLUMN,
     OOBJECT_COLUMN,  /* pointer to the cls, please keep next to last */
     ON_COLUMNS
 };
 
+/** do any special ui initialization. */
+void init_ui (void);
+
 /** The thread used to run gtk-main. 
  */
 void gtkthread (void* data);
@@ -33,9 +37,16 @@ void set_dump_timer (int dumptimer);
 /** Get the current dump timer. */
 int get_dump_timer (void);
 
-/** Set the status text on the main status bar... */
+/** Set the status text on the main status bar, no locking is done, 
+ *  so only do this from the ui-thread ... 
+*/
 void set_status (char* text);
 
+/** Set the status text on the main status bar, locks the ui-structures
+ *  so use this from outside the ui-thread.
+ */
+void set_status_lock (char* text);
+
 /** Check if we have any ui-handling to do. */
 int events_pending (void);
 
